"""Interface to implement necessary methods to work with cache_def."""
from abc import ABCMeta, abstractmethod
from typing import Any


class CacherInterface(metaclass=ABCMeta):
    """Cache Interface for GCP cache system."""

    @abstractmethod
    def get_cache(self, seed: str, cache_hash: str, seconds_expire: int) -> Any:
        """Retrieve the cache based on the received seed and cache_hash."""
        pass

    @abstractmethod
    def set_cache(self, seed: str, cache_hash: str, element: Any) -> bool:
        """Save the cache, with a given seed and cache_hash."""
        pass

    @abstractmethod
    def clear_cache(self, seed: str) -> bool:
        """Remove a specific cache based on a given seed."""
        pass

    @abstractmethod
    def clear_all_cache(self) -> bool:
        """Remove all cache from the cacher."""
        pass



"""Cache on Redis System."""
import logging
from datetime import datetime
from typing import Any

import redis

from .cache_interface import CacherInterface

LOGGER = logging.getLogger(__name__)


class RedisCacher(CacherInterface):
    """Cache Interface for GCP cache system."""

    def __init__(self, host: str = "localhost", port: int = 6379, db: int = None) -> None:
        self.host = host
        self.port = port
        self.db = db
        self.redis = self._get_redis_connection()

    def _get_redis_connection(self) -> redis.Redis:
        """."""
        LOGGER.debug(['RedisCacher', '_get_redis_connection:redishost', self.host])
        return redis.Redis(host=self.host, port=self.port)

    def get_cache(self, seed: str, cache_hash: str, seconds_expire: int) -> Any:
        """Get cache from redis."""
        LOGGER.debug(
            ['RedisCacher', 'get_cache::seed::cache_hast::time', seed, cache_hash, seconds_expire]
        )
        element = self.redis.hget(seed, cache_hash)
        creation_time = self.redis.hget(seed, cache_hash + "::time")
        creation_time = float(creation_time.decode()) if creation_time else None
        now_timestamp = datetime.now().timestamp()

        LOGGER.debug([
            'RedisCacher',
            'get_cache::creation_time::now_timestamp',
            seed,
            creation_time,
            now_timestamp
        ])

        if creation_time:
            if creation_time + seconds_expire > now_timestamp:
                LOGGER.debug(['RedisCacher', 'get_cache::got the cache', seed, cache_hash])
                return element
            else:
                LOGGER.debug(['RedisCacher', 'get_cache::removing old cache', seed, cache_hash])
                self.clear_cache(seed)
        LOGGER.debug(['RedisCacher', 'get_cache::unable to get cache', seed, cache_hash])
        return None

    def set_cache(self, seed: str, cache_hash: str, element: Any) -> bool:
        """Save the cache to redis."""
        now_timestamp = datetime.now().timestamp()
        LOGGER.debug(['redis_cache:set_cache', 'seed::key::time', seed, cache_hash, now_timestamp])

        # Set the cache on MAP of same seed in Redis
        self.redis.hset(seed, cache_hash, element)

        # Set the creation timestamp of cache, to be used as expiration period
        self.redis.hset(seed, cache_hash + "::time", now_timestamp)
        return True

    def clear_cache(self, seed: str, cache_hash: str = None) -> bool:
        """Remove the cache from redis, either all cache from same seed or specific one."""
        if cache_hash:
            LOGGER.debug(['redis_cache:clear_cache', 'seed::cache_hash', seed, cache_hash])
            self.redis.hdel(seed, cache_hash)
        else:
            LOGGER.debug(['redis_cache:clear_cache', 'seed', seed])
            self.redis.delete(seed)
        return True

    def clear_all_cache(self) -> bool:
        """Execute the flushdb command to the redis."""
        return self.redis.flushdb()



"""Decorator for python application cache methods."""

import ast
import hashlib
import logging
import pickle
import pprint
from functools import wraps
from inspect import getcallargs
from typing import Union

from gcp_commons_utils.gcp_cache.cacher import CacherInterface

LOGGER = logging.getLogger(__name__)


# noinspection PyBroadException
class CacheDef:
    """
    Decorator responsible for making a cache.

    Decorator responsible for making a cache of the results of calling a
    method in accordance with the reported.

    :arg seed: Name of the cache
    :arg cacher: The system where the cache will be saved, must be an instance of CacherInterface
    :arg minute_expire: Time in minutes for validity of cache
    :arg debug: Active debug mode
    :arg ftype: Format to store the cache ('pickle', 'literal')
    :arg hash_format: Format of hash when using filesystem
    """

    def __init__(
        self,
        cacher: CacherInterface,
        seed: str = None,
        minute_expire: int = 60 * 24 * 7,
        debug: bool = False,
        ftype: str = 'pickle',
        hash_format: str = '%(md5)s'
    ):
        """Class constructor."""
        LOGGER.debug(
            f"Received variables: seed=[{seed}] - cacher=[{cacher}] - minute_expire=["
            f"{minute_expire}] - debug=[{debug}] - ftype=[{ftype}] - hash_format=[{hash_format}]"
        )

        self.seed = seed
        self.cacher = cacher
        self.seconds_expire = minute_expire * 60 if minute_expire >= 1 else 60 * 60 * 24 * 7
        self.debug = debug
        self.ftype = ftype
        self.hash_format = hash_format
        self._config = {}

    @property
    def config(self):
        """Getter for config property."""
        if not self._config:
            self._config = {
                "seed": self.seed,
                "cacher": self.cacher,
                "seconds_expire": self.seconds_expire,
                "debug": self.debug,
                "ftype": self.ftype,
                "hash_format": self.hash_format,
            }
            LOGGER.debug("Setting values for config. [%s]" % self._config)
        LOGGER.debug("Returning values for config. [%s]" % self._config)
        return self._config

    def __call__(self, call, *args, **kwargs):
        """Intercept the all calls for the cached method."""
        LOGGER.info("Call for cache_def.")
        config = self.config.copy()

        @wraps(call)
        def new_def(*in_args, **in_kwargs):

            if self.cacher is None:
                LOGGER.warning("CacheDef::cache_def: WARNING NO CACHE IS BEING USED!")
                return call(*in_args, **in_kwargs)

            if self.seed is None:
                self.seed = call.__qualname__.replace("<", "").replace(">", "").replace(".", "::")
                self.seed += f"::{int(self.seconds_expire / 60)}"
            resp = None

            call_kwargs = getcallargs(call, *in_args, **in_kwargs)
            if 'self' in call_kwargs:
                call_kwargs.pop('self')
            cache_hash = self._generate_hash(config=config, kwargs=call_kwargs)
            cache_key = 'cache:' + cache_hash

            if not kwargs.get('renew_cache'):
                resp = self.cacher.get_cache(self.seed, cache_key, self.seconds_expire)
                resp = self.loads(resp)
                LOGGER.info(['sbox_cache', 'get cache', self.seed])

            if resp is None:
                LOGGER.info(['sbox_cache', 'no cache', self.seed])

                # Make the normal call to the function
                resp = call(*in_args, **in_kwargs)

                # Set the cache
                # thread BIG QUERY
                self.cacher.set_cache(self.seed, cache_key, self.dumps(resp))
                LOGGER.info(['sbox_cache', 'save cache', self.seed])
            return resp
        return new_def

    def loads(self, element):
        """Serialize the element retrieved from cacher."""
        if element is None:
            return element
        if self.ftype == 'pickle':
            return pickle.loads(element)
        if self.ftype == 'literal':
            return ast.literal_eval(element)
        raise TypeError(f"Type of {self.ftype} not supported.")

    def dumps(self, element):
        """Deserialize the element to save in the cacher."""
        if self.ftype == 'pickle':
            return pickle.dumps(element)
        if self.ftype == 'literal':
            r = pprint.pformat(element).encode('utf-8')
            return r
        raise TypeError(f"Type of {self.ftype} not supported.")

    def _generate_hash(self, config, kwargs):
        LOGGER.info(f"GENERATING CACHE_HASH....")
        LOGGER.debug(['sbox_cache', '_gera_hash::kwargs', kwargs])

        if 'renew_cache' in kwargs:
            kwargs.pop('renew_cache')

        s = self.seed + str(self.seconds_expire) + pprint.pformat([kwargs])

        if hasattr(s, 'encode'):
            s = s.encode('utf-8')

        hash_format = config.get('hash_format')
        hash_key = hash_format % dict(md5=hashlib.md5(s).hexdigest(), **kwargs)

        LOGGER.debug(['sbox_cache', '_gera_hash', hash_key])
        LOGGER.debug(f"CACHE_HASH = {hash_key}")
        return hash_key


def cache_def(
    cacher: Union[CacherInterface, None],
    seed: str = None,
    minute_expire: int = 60,
    debug: bool = False,
    ftype: str = 'pickle',
    hash_format: str = '%(md5)s'
):
    """
    Decorator responsible for making a cache.

    Decorator responsible for making a cache of the results
    of calling a method in accordance with the reported.

    :arg seed: Name of the cache
    :arg cacher: The system where the cache will be saved, must be an instance of CacherInterface
    :arg minute_expire: Time in minutes for validity of cache
    :arg debug: Active debug mode
    :arg ftype: Format to store the cache ('pickle', 'literal')
    :arg hash_format: Format of hash when using filesystem
    """
    return CacheDef(
        cacher=cacher,
        seed=seed,
        minute_expire=minute_expire,
        debug=debug,
        ftype=ftype,
        hash_format=hash_format,
    )


def clear_cache(seed: str, cacher: CacherInterface):
    """Remove the cache based on it seed."""

    return cacher.clear_cache(seed)


def clear_all_cache(cacher: CacherInterface) -> bool:
    """Remove all cache from the given gacher."""

    return cacher.clear_all_cache()



import base64

import binascii
import json

class DecodeUtil:

    def decode_input_to_dict(input_request):
        """Decode input request to dict python."""
        if isinstance(input_request, dict):
            return input_request

        if isinstance(input_request, str):
            input_request = input_request.encode()

        try:
            input_request = base64.decodebytes(input_request)
        except (TypeError, binascii.Error):
            pass

        try:
            input_request = json.loads(input_request)
        except Exception:
            raise Exception(f"Invalid JSON object. {input_request} ...")
        return input_request



"""Component test for cache_def."""
import time
from gcp_commons_utils.gcp_cache import cache_def, clear_cache
from gcp_commons_utils.gcp_cache.cacher import RedisCacher


def test_call_on_cache_def():
    """Test if method are called when cache does not exists."""
    start = time.time()
    redis_caching = RedisCacher(host="127.0.0.1")
    seed = "testing"
    clear_cache(seed, redis_caching)

    @cache_def(cacher=redis_caching, minute_expire=1)
    def cache_test(a, b):
        time.sleep(5)
        return a * b

    resp = cache_test(2, 4)
    duration = (time.time() - start)
    assert resp == 8
    assert duration >= 5


def test_call_multiple_times():
    """Test if cache are valid when calling the method multiple times and in different ways."""
    start = time.time()
    redis_caching = RedisCacher(host="127.0.0.1")
    seed = "testing"
    clear_cache(seed, redis_caching)

    @cache_def(seed="testing", cacher=redis_caching, minute_expire=1)
    def cache_test(a, b):
        time.sleep(5)
        return a * b

    resp = (
        cache_test(2, 4),
        cache_test(a=2, b=4),
        cache_test(2, b=4),
        cache_test(b=4, a=2)
    )
    duration = (time.time() - start)

    assert resp == (8, 8, 8, 8)
    assert duration <= 7


def test_multiple_call_different_param():
    """Test if cache are different for method when parameters are different."""
    start = time.time()
    redis_caching = RedisCacher(host="127.0.0.1")
    seed = "testing"
    clear_cache(seed, redis_caching)

    @cache_def(seed="testing", cacher=redis_caching, minute_expire=1)
    def cache_test(a, b):
        time.sleep(5)
        return a * b

    resp = (
        cache_test(a=2, b=4),
        cache_test(2, b=7),
        cache_test(2, 8),

        cache_test(2, 4),
        cache_test(b=7, a=2),
        cache_test(a=2, b=8),
    )
    duration = (time.time() - start)

    assert resp == (8, 14, 16, 8, 14, 16)
    assert duration >= 15 <= 17


def test_clear_cache():
    """Test if cache are cleared."""
    start = time.time()
    redis_caching = RedisCacher(host="127.0.0.1")
    seed = "testing"
    clear_cache(seed, redis_caching)

    @cache_def(seed="testing", cacher=redis_caching, minute_expire=1)
    def cache_test(a, b):
        time.sleep(5)
        return a * b

    resp = cache_test(a=2, b=4)
    assert resp == 8

    clear_cache(seed, redis_caching)

    resp = cache_test(a=2, b=4)
    assert resp == 8

    duration = (time.time() - start)

    assert duration >= 10


def test_none_cacher():
    """Test if cacher is None, no cache will be used."""
    start = time.time()
    seed = "testing"

    @cache_def(seed=seed, cacher=None, minute_expire=1)
    def cache_test(a, b):
        time.sleep(5)
        return a * b

    resp = (
       cache_test(a=2, b=5),
       cache_test(2, 5)
    )

    duration = (time.time() - start)

    assert resp == (10, 10)
    assert duration >= 10
